//This routine is based on zippie.cpp
#include <stdio.h>
#include <stdlib.h>
#include <list>
#include <math.h>
#include <string.h>
#include "idl_export.h"
#include "idl_tools.h"

#include "../ZipArchive/ZipArchive.h"
#include "../ZipArchive/ZipPlatform.h"
#include "../ZipArchive/ZipException.h"
#include <io.h>
#include "Str.h"

//using namespace std;

ZIPSTRINGCOMPARE pZipComp;

struct CZipAddFileInfo
{
	CZipAddFileInfo(const CZipString& szFilePath, const CZipString& szFileNameInZip)
		:m_szFilePath(szFilePath),	m_szFileNameInZip(szFileNameInZip)
	{
		int is = szFileNameInZip.GetLength();
		m_iSeparators = 0;
		for (int i = 0; i < is; i++)
			if (CZipPathComponent::IsSeparator(szFileNameInZip[i]))
				m_iSeparators++;
	}
	CZipString m_szFilePath, m_szFileNameInZip;
	
	bool CheckOrder(const CZipString& sz1, const CZipString& sz2, 
		int iSep1, int iSep2, bool bCheckTheBeginning = false) const
	{
		if (iSep1)
		{
			if (iSep2)
			{
				if (iSep1 == iSep2)
					return (sz1.*pZipComp)(sz2) < 0;

				if (bCheckTheBeginning)
				{

					int is = sz1.GetLength() > sz2.GetLength() ? sz2.GetLength() : sz1.GetLength();
					int iSeparators = 0;
					// find the common path beginning
					int iLastSepPos = -1;
					for (int i = 0; i < is; i++)
					{
						CZipString sz = sz2.Mid(i, 1);
						if ((sz1.Mid(i, 1).*pZipComp)(sz) != 0) // must be Mid 'cos of case sens. here
							break;
						else if (CZipPathComponent::IsSeparator(sz[0]))
						{
							iLastSepPos = i;
							iSeparators++;
						}
					}
					
					// if necessary remove the common path beginning and check again
					if (iLastSepPos != -1)
						return CheckOrder(sz1.Mid(iLastSepPos), sz2.Mid(iLastSepPos), iSep1 - iSeparators, iSep2 - iSeparators);
							
				}
				return (sz1.*pZipComp)(sz2) < 0;		
			} 
			else
				return false;
		}
		else
			if (iSep2)
				return true;
			else
				return (sz1.*pZipComp)(sz2) < 0;
	}
	bool operator>(const CZipAddFileInfo& wz) const
	{
		bool b = CheckOrder(m_szFileNameInZip, wz.m_szFileNameInZip,
			m_iSeparators, wz.m_iSeparators, true);
		return b;
	}
protected:
	int m_iSeparators; // for a sorting puroposes
};

/*globals here*/
static char statusBuffer[256] ;
static char errorBuffer[256] ;



/* function protos */
extern IDL_VPTR IDL_CDECL zipList(int argc, IDL_VPTR argv[]);
void IDL_CDECL zipExtract(int argc, IDL_VPTR argv[], char *argk);
void IDL_CDECL zipCompress(int argc, IDL_VPTR argv[], char *argk);

/* define the functions */
static IDL_SYSFUN_DEF2 IDLtoZipfile_functions[] = {
	{(IDL_FUN_RET)zipList, "ZIPLIST",   1, 1, 0, 0},
};

static IDL_SYSFUN_DEF2 IDLtoZipfile_procedures[] = {
	{(IDL_FUN_RET) zipCompress, "ZIPCOMPRESS", 2, 2, IDL_SYSFUN_DEF_F_KEYWORDS, 0},
	{(IDL_FUN_RET) zipExtract, "ZIPEXTRACT", 1, 2, IDL_SYSFUN_DEF_F_KEYWORDS, 0},
};

/* startup call when DLM is loaded */
int zipfile_startup(void)
{
        if (!IDL_SysRtnAdd(IDLtoZipfile_functions, TRUE, ARRLEN(IDLtoZipfile_functions))) 
		{
                return IDL_FALSE;
        }
        if (!IDL_SysRtnAdd(IDLtoZipfile_procedures, FALSE, ARRLEN(IDLtoZipfile_procedures))) 
		{
               return IDL_FALSE;
        }

/*Register the exit handler*/
        IDL_ExitRegister(zipfile_exit_handler);
        return(IDL_TRUE);
}

/* called when IDL is shutdown */
void zipfile_exit_handler(void)
{

}

/*{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|*/

void FindInZip(CZipArchive& zip, std::list<CZipString>& l, CZipIndexesArray& n)
{

	for (std::list<CZipString>::iterator it = l.begin(); it != l.end(); ++it)
		zip.FindMatches(*it, n);
}

/*{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|*/

IDL_VPTR IDL_CDECL zipList(int argc, IDL_VPTR argv[])
{
	/*
     Called in IDL as 
	 filesInZip = zipList(inputZipFile)
	*/ 

	try
	{
	 IDL_ENSURE_STRING(argv[0]);

     CZipArchive zip;	
	 CZipString szArchive;

	 szArchive = _T(IDL_STRING_STR(&(argv[0]->value.str)));

	 CZipPathComponent zpc(szArchive);

	 if (zpc.GetFileExt().IsEmpty())
		 szArchive.Insert(szArchive.GetLength(),_T(".zip"));

	 int iVolumeSize = 0;
     int iMode = CZipArchive::zipOpen;
	 zip.Open(szArchive, iMode, iVolumeSize);
     ZIP_INDEX_TYPE iCount = zip.GetCount();
	 //make a string array of the right size

 	 IDL_STRING *pisArray;
	 IDL_LONG lDimensions[IDL_MAX_ARRAY_DIM];
	 IDL_VPTR ivReturn;

	 lDimensions[0] = iCount;
	 pisArray = (IDL_STRING *)IDL_MakeTempArray((int) IDL_TYP_STRING,
	 		     1, lDimensions, IDL_ARR_INI_NOP, &ivReturn);

	 for (ZIP_INDEX_TYPE i = 0; i < iCount; i++) {
	    CZipFileHeader fh;
		if (zip.GetFileInfo(fh, i)) {
			IDL_StrStore(&(pisArray[i]),  (char *)((LPCTSTR)fh.GetFileName()) );
			IDL_StrEnsureLength(&(pisArray[i]), fh.GetFileName().GetLength());  //garbage at the end if we don't do this
		} else {
			zip.Close(true);
			IDL_Deltmp(ivReturn);
            CZipException::Throw(CZipException::badZipFile);		}
	  }

	  zip.Close();

      return ivReturn;

    }
	catch(CZipException& e)
	{
		sprintf(statusBuffer, _T("Error while processing archive %s\n"), (LPCTSTR)e.GetErrorDescription());
		IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_RET, statusBuffer);
		return IDL_GettmpLong(0);
	}

}

//check idltocanon project for keyword examples

/*{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|*/

void IDL_CDECL zipExtract(int argc, IDL_VPTR argv[], char *argk)
{
	/*
     Called in IDL as 
	 zipExtract, inputZipFile, [fileList], path=path
	 where path is the path to exract to
	*/ 

	 typedef struct {
      IDL_KW_RESULT_FIRST_FIELD; 
	  IDL_STRING path; int pathThere;
	} KW_RESULT;

	static  IDL_KW_PAR      kw_pars[] = { IDL_KW_FAST_SCAN,
		{"PATH",IDL_TYP_STRING,1,0,(int *)IDL_KW_OFFSETOF(pathThere),(char *)IDL_KW_OFFSETOF(path)},
		{NULL}
        };

	KW_RESULT kw;
	argc = IDL_KWProcessByOffset(argc,argv,argk,kw_pars,(IDL_VPTR *)0,1,&kw);

    //make sure the filename is a string
    IDL_ENSURE_STRING(argv[0]);
    
    //check to see if we have a path to extract to
	CZipString szPath;
	if (kw.pathThere) {	
		szPath = IDL_STRING_STR(&kw.path);
	}

     CZipArchive zip;	
	 CZipString szArchive;

	 szArchive = _T(IDL_STRING_STR(&(argv[0]->value.str)));

	 CZipPathComponent zpc(szArchive);

	 if (zpc.GetFileExt().IsEmpty())
		 szArchive.Insert(szArchive.GetLength(),_T(".zip"));

	 int iVolumeSize = 0;
     int iMode = CZipArchive::zipOpen;
	 //no case sensitivity
	 zip.SetCaseSensitivity(false);

	 zip.Open(szArchive, iMode, iVolumeSize);
     ZIP_INDEX_TYPE iCount = zip.GetCount();

	 CZipIndexesArray vRevised;
	 if (argc == 1) { //extract all the files in the zip
			ZIP_INDEX_TYPE iMax = zip.GetCount();
			for (ZIP_INDEX_TYPE i = 0; i < iMax; i++)
				vRevised.Add(i);	
	 } else 	{
			CZipIndexesArray numbers;
			IDL_STRING *pIDLstringDesc;

			/*has to be a string and an array.*/
			IDL_ENSURE_ARRAY(argv[1]);
			IDL_ENSURE_STRING(argv[1]);
			/*set a pointer to the first IDL_STRING in the IDL_ARRAY*/
			pIDLstringDesc = (IDL_STRING *)argv[1]->value.arr->data;
	
			int iCount = argv[1]->value.arr->n_elts;
			
			if (iCount > 0)
			{
				std::list<CZipString> lFiles;
				for (int i = 0; i < iCount; i++) {
					CStr tempString = IDL_STRING_STR(&(pIDLstringDesc[i]));
					/*have to replace \ with \\ */
	                tempString.Replace("\\","\\\\");
					lFiles.push_back(tempString.Buffer()  );
				}
				FindInZip(zip, lFiles, numbers);
			}

			ZIP_INDEX_TYPE iMax = zip.GetCount();
			if (iMax > 0)
			{
				iMax--;
				for (ZIP_ARRAY_SIZE_TYPE i = 0; i < numbers.GetSize(); ++i)
				{
					ZIP_INDEX_TYPE x = numbers[i];
					if (x > iMax)
						continue;
					bool bNew = true;
					for (ZIP_ARRAY_SIZE_TYPE j = 0; j < vRevised.GetSize(); ++j)
						if (vRevised[j] == numbers[i])
						{
							bNew = false;
							break;
						}
					if (bNew)
						vRevised.Add(x);
				}
			}
	 }

	 for (ZIP_ARRAY_SIZE_TYPE k = 0; k < vRevised.GetSize(); ++k) {
		ZIP_INDEX_TYPE iFile = vRevised[k];
		bool ok;
		try
		{
			ok = zip.ExtractFile(iFile, szPath, true);
		}
		catch (...)
		{					
			ok = false;
		}
		CZipFileHeader fh;
		if (ok)
		{
			if (zip.GetFileInfo(fh, iFile)){
				sprintf(statusBuffer, _T("%s extracted\n"), (LPCTSTR)fh.GetFileName());
				IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_RET, statusBuffer);
			}
		} else 	{
			if (zip.GetFileInfo(fh, iFile))
					sprintf(statusBuffer, _T("Error while extracting file %s\n"), (LPCTSTR)fh.GetFileName());
				else
					sprintf(statusBuffer, _T("Error while extracting a file. There are troubles with getting info from file number %d\n"), iFile);

			IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_RET, statusBuffer);
			break;
		}
	}	 

	 IDL_KW_FREE;
}

/*{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|{{:|*/

void IDL_CDECL zipCompress(int argc, IDL_VPTR argv[], char *argk)
{
	/*
     Called in IDL as 
	 zipCompress, outputZipFile, [fileList], update=update, compress=[1-9]
	 update = flag for adding files to existing zip file
	 compress = 1 to 9 for compression level

	 no wildcards allowed in fileList, user must specify each one
	*/ 

	 typedef struct {
      IDL_KW_RESULT_FIRST_FIELD; 
	  IDL_LONG compress;
	  IDL_LONG subdirs;
	  IDL_LONG update;
	} KW_RESULT;

	 static  IDL_KW_PAR      kw_pars[] = { IDL_KW_FAST_SCAN,
		{"COMPRESS",IDL_TYP_LONG,1,IDL_KW_ZERO,0, (char *)IDL_KW_OFFSETOF(compress)},
		{"UPDATE",IDL_TYP_LONG,1,IDL_KW_ZERO,0, (char *)IDL_KW_OFFSETOF(update)},
		{NULL}
        };

	 KW_RESULT kw;
	 argc = IDL_KWProcessByOffset(argc,argv,argk,kw_pars,(IDL_VPTR *)0,1,&kw);

     //make sure the filename is a string
     IDL_ENSURE_STRING(argv[0]);
     IDL_ENSURE_ARRAY(argv[1]); //array only

	 bool bUpdateMode=false;
     if (kw.update) bUpdateMode=true;

     //no recursion allowed
	 bool bRecursive = false;
     //hardwire full path
	 bool bFullPath = true;

	 int iLevel = 5;
	 if (kw.compress) iLevel = kw.compress;

     CZipArchive zip;	
	 CZipString szArchive;

	 szArchive = _T(IDL_STRING_STR(&(argv[0]->value.str)));

	 CZipPathComponent zpc(szArchive);

	 if (zpc.GetFileExt().IsEmpty())
		 szArchive.Insert(szArchive.GetLength(),_T(".zip"));

	 int iVolumeSize = 0;
     int iMode = CZipArchive::zipOpen;

     if (!bUpdateMode || !ZipPlatform::FileExists(szArchive))
		iMode = CZipArchive::zipCreate;

	 int iSmartLevel = CZipArchive::zipsmSafeSmart;

     //no case sensitivity
	 zip.SetCaseSensitivity(false);
 
	 //can have a failure if we don't have a correct filename
	try {
	   zip.Open(szArchive, iMode, iVolumeSize);
	 } catch (...) {					
		 sprintf(statusBuffer, _T("Error opening %s\n"), (LPCTSTR)szArchive);
		 IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_RET, statusBuffer);
		 zip.Close();
		 return;
	}
	CZipIndexesArray numbers;
	IDL_STRING *pIDLstringDesc;

	/*has to be a string and an array.*/
	IDL_ENSURE_ARRAY(argv[1]);
	IDL_ENSURE_STRING(argv[1]);
	/*set a pointer to the first IDL_STRING in the IDL_ARRAY*/
	pIDLstringDesc = (IDL_STRING *)argv[1]->value.arr->data;

	int iCount = argv[1]->value.arr->n_elts;
	
	std::list<CZipString> lFiles;
	for (int i = 0; i < iCount; i++) {
			CStr tempString = IDL_STRING_STR(&(pIDLstringDesc[i]));
			/*have to replace \ with \\ */
            //tempString.Replace("\\","\\\\");
     		lFiles.push_back(tempString.Buffer()  );
	}

	std::list<struct CZipAddFileInfo> rev;

	//ZipArchive supports excluding files, don't support it yet but create varialbe so that 
	//we match code in zippie.cp

	std::list<CZipString> excl;

	for (std::list<CZipString>::iterator it = lFiles.begin(); it != lFiles.end(); ++it)				
	{
		// that is how the filename will look in the archive
		CZipString sz = zip.PredictFileNameInZip(*it, bFullPath);
    	if (!sz.IsEmpty())
		{
			bool bAdd = true;
			for (std::list<CZipString>::iterator itt = excl.begin(); itt != excl.end(); ++itt)
			{
				if (!((*itt).*pZipComp)(*it))
				{
					bAdd = false;
					break;
				}
			}
			if (bAdd)
				rev.push_back(CZipAddFileInfo(*it, sz));
			}
  	     }
	 lFiles.clear();

	 for (std::list<struct CZipAddFileInfo>::iterator it1 = rev.begin(); it1 != rev.end(); ++it1)
	 {	
		if (zip.AddNewFile((*it1).m_szFilePath, iLevel, bFullPath, iSmartLevel))
		{
			sprintf(statusBuffer,_T("%s added\n"), (LPCTSTR)(*it1).m_szFileNameInZip);
		}
		else sprintf(statusBuffer,_T("%s not added\n"), (LPCTSTR)(*it1).m_szFilePath);
        IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_RET, statusBuffer);
	 }
           
     zip.Close();

	 IDL_KW_FREE;
}